import numpy as np
import scipy as sp

#kernel paramter, only support RBF or ARD kernel now
#k(x,y) = w * exp(-(x-y)^\top diag(gamma) (x-y)) + kreg*I(x==y)
class KernelParam:
    #note: "ard" needs gamma to be vector
    def __init__(self, type='rbf', w=1, gamma=1, kreg=1e-5):
        self.type = type
        self.w = w;
        self.gamma = gamma
        self.kreg = kreg
    def recover(self, type, params):
        self.type = type
        self.w = params[0]
        self.kreg = params[1]
        self.gamma = params[2:]
    def tovec(self):
        params = np.array([self.w, self.kreg])
        return np.append(params, self.gamma)
        
           
#all input are assumped to be np.matrix        
#kernel function of a n*p matrix, return a n*n Gram matrix
def kernel_func(X, ker_param):
    X = X.A
    n,p = X.shape
    K = None
    if ker_param.type == 'rbf':
        norm2 = np.sum(X**2,axis=1)
        norm2 = np.tile(norm2, [n,1])
        K = norm2 + norm2.T- 2*np.dot(X, X.T)
        K = ker_param.w*np.exp(-param.gamma*K)
    elif ker_param.type == 'ard':
        Y = np.dot(X, np.diagflat(np.sqrt(ker_param.gamma)))
        norm2 = np.sum(Y**2,axis=1)
        norm2 = np.tile(norm2, [n,1])
        K = norm2 + norm2.transpose() - 2*np.dot(Y, Y.transpose())
        K = ker_param.w*np.exp(-K)
    else:
        print 'not supported kernel type'
        sys.exit(1)

    K = K + np.eye(n)*ker_param.kreg
    return np.mat(K)

#cross kernel function between X and Y (n*p, m*p), return n*m matrix
def kernel_cross_func(X, Y, ker_param):
    X = X.A
    Y = Y.A
    n = X.shape[0]
    m = Y.shape[0]
    if ker_param.type == 'rbf':
        P = np.sum(X**2, axis=1).reshape([n,1])
        Q = np.sum(Y**2, axis=1).reshape([m,1])
        K = ker_param.w*np.exp(-ker_param.gamma*(np.tile(P,[1,m]) + np.tile(Q.T,[n,1]) - 2*np.dot(X,Y.T)))
    elif ker_param.type == 'ard':
        sqrt_ga = np.diagflat(np.sqrt(ker_param.gamma))
        X1 = np.dot(X, sqrt_ga)
        Y1 = np.dot(Y, sqrt_ga)
        P = np.sum(X1**2, axis=1).reshape([n,1])
        Q = np.sum(Y1**2, axis=1).reshape([m,1])
        K = ker_param.w*np.exp(-(np.tile(P,[1,m]) + np.tile(Q.T,[n,1]) - 2*np.dot(X1,Y1.T)))
    else:
        print 'not supported kernel type'
        sys.exit(1)

    return np.mat(K)


#kernel gradient
#gradient for tr(A\cdot dK), K = k(X,X), now only support kernel param gradient for ARD
def ker_grad(X, A, K, ker_param, calculate_gradient_kernel_params=True): 
    X = X.A
    A = A.A
    K = K.A
    n,d = X.shape
    if ker_param.type == 'ard' or ker_param.type == 'rbf':
        K = K - ker_param.kreg*np.eye(n)        
        AAK = (A + A.T)*K
        if calculate_gradient_kernel_params:
            AK = A.T*K
    else:
        print 'not supported kernel type'
        sys.exit(1)

    g = np.zeros(X.shape)
    if calculate_gradient_kernel_params:
        gkp = np.zeros(d + 1)
    
    for j in range(d):
        repU = np.tile(X[:,j].reshape([n,1]), [1,n])
        if ker_param.type == 'ard':
            B = repU.T - repU
            # -2*kernel.gamma(j)*sum(AAK.*B)';
            g[:,j] = -2*ker_param.gamma[j]*np.sum(AAK*B, axis=0)
            if calculate_gradient_kernel_params:
                gkp[j] = -np.sum(AK*(B.T**2))
        elif ke_param.type == 'rbf':
            B = repU.T - repU
            g[:,j] = -2*ker_param.gamma*np.sum(AAK*B, axis=0)

    if ker_param.type == 'ard' and calculate_gradient_kernel_params:
        gkp[d] = np.sum(AK)/ker_param.w

    if calculate_gradient_kernel_params:
        return [np.mat(g), np.mat(gkp.reshape([gkp.size,1]))]
    else:
        return np.mat(g)


#gradient for kcross kernels, tr(A * dK_{X,Y}), the gradient is over X
def ker_cross_grad(X, Y, A, K, ker_param, calculate_gradient_kernel_params=True):
    n,d = X.shape
    m = Y.shape[0]
    AK = np.multiply(A.T, K)
    g = np.mat(np.zeros(X.shape))
    if calculate_gradient_kernel_params and ker_param.type == 'ard':
        gkp = np.mat(np.zeros([d+1,1]))
        #print AK[0,:]
    for j in xrange(d):
        repX = np.tile(X[:,j], [1, m])
        repY = np.tile(Y[:,j], [1, n])
        if ker_param.type == 'ard':
            B = repX - repY.T
            AKB = np.multiply(AK, B)
            g[:,j] = -2*ker_param.gamma[j]*np.sum(AKB, 1)
            if calculate_gradient_kernel_params:
                gkp[j,0] = -np.sum(np.multiply(AKB, B))
                '''
                if j==0:
                    AKBB = np.multiply(AKB, B)
                    print AKBB.shape
                    print np.sum(AKBB,1)
                    print gkp[j,0]
                '''
                
        elif ker_param.type == 'rbf':
            B = repX - repY.T
            AKB = np.multiply(AK,B)
            g[:,j] = -2*ker_param.gamma*np.sum(AKB, 1)
        else:
            print 'not supported kernel type'
            sys.exit(1)
    
    if calculate_gradient_kernel_params and ker_param.type == 'ard':
        gkp[-1,0] = np.sum(AK)/ker_param.w
        return [g, gkp]
    else:
        return g

   


#for test
if __name__ == '__main__':
    ''' test group 1'''
    X = np.mat(np.array([1,2,3,4,5,6]).reshape(3,2))
    ker_param = KernelParam(type = 'ard', w = 1, gamma = np.array([1,1]))
    K = kernel_func(X, ker_param) 
    A = np.mat(np.array([1,2,3,4,5,6,7,8,9]).reshape(3,3))
    [g, gkp] = ker_grad(X, A, K, ker_param)
    A = np.mat(np.array([1,2,3,4,5,6,7,8,9,10,11,12]).reshape(4,3))
    Y = np.mat(np.array([8,7,6,5,4,3,2,1]).reshape(4,2))
    K = kernel_cross_func(X,Y, ker_param)
    [g, gkp] = ker_cross_grad(X,Y,A,K, ker_param)

    ''' test group 2 '''
    X = np.mat(np.random.randn(4,3)*0.1)
    Y = np.mat(np.random.randn(3,3)*0.1)
    A = np.mat(np.random.randn(4,4)*0.1)
    ker_param = KernelParam(type='ard', w = 0.2, gamma = np.array([0.1, 2.3, 3.4]), kreg=0.8)
    K = kernel_func(X, ker_param)
    [g,gkp] = ker_grad(X, A, K, ker_param)
    K = kernel_cross_func(X, Y, ker_param)
    A = np.mat(np.random.randn(3,4)*0.1)
    [g, gkp] = ker_cross_grad(X, Y, A, K, ker_param) 

    '''
    print ker_param.tovec()
    ker_param.recover('ard', np.array([1,2,0.1,0.1,0.1]))
    print ker_param.type
    print ker_param.w, ker_param.kreg
    print ker_param.gamma
    '''

